"""MuMoT controller classes."""

import base64

from IPython.display import Javascript, display
import ipywidgets.widgets as widgets
from ipywidgets import HTML
from sympy import Symbol

from . import (
    consts,
    defaults,
    exceptions,
    utils,
    views,
)


class MuMoTcontroller:
    """Controller for a model view."""

    _view = None
    # dictionary of LaTeX labels for widgets, with parameter name as key
    _paramLabelDict = None
    # dictionary of controller widgets only for the free parameters of the model, with parameter name as key
    _widgetsFreeParams = None
    # dictionary of controller widgets for the special parameters (e.g., simulation length, initial state), with parameter name as key
    _widgetsExtraParams = None
    # dictionary of controller widgets, with parameter that influence only the plotting and not the computation
    _widgetsPlotOnly = None
    # list keeping the order of the extra-widgets (_widgetsExtraParams and _widgetsPlotOnly)
    _extraWidgetsOrder = None
    # replot function widgets have been assigned (for use by MuMoTmultiController)
    _replotFunction = None
    # redraw function widgets have been assigned (for use by MuMoTmultiController)
    _redrawFunction = None
    # widget for simple error messages to be displayed to user during interaction
    _errorMessage = None
    # plot limits slider widget
    _plotLimitsWidget = None  # @todo: is it correct that this is a variable of the general MuMoTcontroller?? it might be simply added in the _widgetsPlotOnly dictionary
    # system size slider widget
    _systemSizeWidget = None
    # bookmark button widget
    _bookmarkWidget = None
    # advanced tab widget
    _advancedTabWidget = None
    # download button widget
    _downloadWidget = None
    # download link widget
    _downloadWidgetLink = None

    def __init__(self, paramValuesDict, paramLabelDict=None,
                 continuousReplot: bool = False, showPlotLimits: bool = False,
                 showSystemSize: bool = False, advancedOpts=None, **kwargs) -> None:
        self._silent = kwargs.get('silent', False)
        self._paramLabelDict = paramLabelDict if paramLabelDict is not None else {}
        self._widgetsFreeParams = {}
        self._widgetsExtraParams = {}
        self._widgetsPlotOnly = {}
        self._extraWidgetsOrder = []

        for paramName in sorted(paramValuesDict.keys()):
            if paramName == 'plotLimits' or paramName == 'systemSize':
                continue
            if not paramValuesDict[paramName][-1]:
                paramValue = paramValuesDict[paramName]
                widget = widgets.FloatSlider(value=paramValue[0], min=paramValue[1],
                                             max=paramValue[2], step=paramValue[3],
                                             readout_format='.' + str(utils._count_sig_decimals(str(paramValue[3]))) + 'f',
                                             description=r'\(' + utils._doubleUnderscorify(utils._greekPrependify(self._paramLabelDict.get(paramName, paramName))) + r'\)',
                                             style={'description_width': 'initial'},
                                             continuous_update=continuousReplot)
                self._widgetsFreeParams[paramName] = widget
                if not self._silent:
                    display(widget)
        if showPlotLimits:
            if not paramValuesDict['plotLimits'][-1]:
                paramValue = paramValuesDict['plotLimits']
                self._plotLimitsWidget = widgets.FloatSlider(
                    value=paramValue[0], min=paramValue[1],
                    max=paramValue[2], step=paramValue[3],
                    readout_format='.' + str(utils._count_sig_decimals(str(paramValue[3]))) + 'f',
                    description='Plot limits',
                    style={'description_width': 'initial'},
                    continuous_update=continuousReplot)
                # @todo: it would be better to remove self._plotLimitsWidget and use the self._widgetsExtraParams['plotLimits'] = widget
                if not self._silent:
                    display(self._plotLimitsWidget)

        if showSystemSize:
            if not paramValuesDict['systemSize'][-1]:
                paramValue = paramValuesDict['systemSize']
                self._systemSizeWidget = widgets.FloatSlider(
                    value=paramValue[0], min=paramValue[1],
                    max=paramValue[2], step=paramValue[3],
                    readout_format='.' + str(utils._count_sig_decimals(str(paramValue[3]))) + 'f',
                    description='System size',
                    style={'description_width': 'initial'},
                    continuous_update=continuousReplot)
                if not self._silent:
                    display(self._systemSizeWidget)

        # create advanced widgets (that will be added into the 'Advanced options' tab)
        initialState = self._createAdvancedWidgets(advancedOpts, continuousReplot)
        self._orderAdvancedWidgets(initialState)
        # add widgets to the Advanced options tab
        if not self._silent:
            self._displayAdvancedOptionsTab()

        self._bookmarkWidget = widgets.Button(description='',
                                              disabled=False,
                                              button_style='',
                                              tooltip='Paste bookmark to log',
                                              icon='fa-bookmark')
        self._bookmarkWidget.on_click(self._print_standalone_view_cmd)
        bookmark = kwargs.get('bookmark', True)

        self._downloadWidget = widgets.Button(description='',
                                              disabled=False,
                                              button_style='',
                                              tooltip='Download results',
                                              icon='fa-save')
        self._downloadWidgetLink = HTML(self._create_download_link("", "", ""),
                                        visible=False)
        self._downloadWidgetLink.layout.visibility = 'hidden'
        self._downloadWidget.on_click(self._download_link_unsupported)

        if not self._silent and bookmark:
            # display(self._bookmarkWidget)

            box_layout = widgets.Layout(display='flex',
                                        flex_flow='row',
                                        align_items='stretch',
                                        width='70%')
            threeButtons = widgets.Box(children=[self._bookmarkWidget,
                                                 self._downloadWidget,
                                                 self._downloadWidgetLink],
                                       layout=box_layout)
            display(threeButtons)

        widget = widgets.HTML()
        widget.value = ''
        self._errorMessage = widget
        if not self._silent and bookmark:
            display(self._errorMessage)

    def _print_standalone_view_cmd(self, _includeParams) -> None:
        self._errorMessage.value = "Pasted bookmark to log - view with showLogs(tail = True)"
        self._view._print_standalone_view_cmd(True)

    # set the functions that must be triggered when the widgets are changed.
    # @param[in]    recomputeFunction    The function to be called when recomputing is necessary
    # @param[in]    redrawFunction    The function to be called when only redrawing (relying on previous computation) is sufficient
    def _setReplotFunction(self, recomputeFunction, redrawFunction=None) -> None:
        """set the functions that must be triggered when the widgets are changed.
        :param recomputeFunction
            The function to be called when recomputing is necessary
        :param redrawFunction
            The function to be called when only redrawing (relying on previous computation) is sufficient"""
        self._replotFunction = recomputeFunction
        self._redrawFunction = redrawFunction
        for widget in self._widgetsFreeParams.values():
            # widget.on_trait_change(recomputeFunction, 'value')
            widget.observe(recomputeFunction, 'value')
        for widget in self._widgetsExtraParams.values():
            widget.observe(recomputeFunction, 'value')
        if self._plotLimitsWidget is not None:
            self._plotLimitsWidget.observe(recomputeFunction, 'value')
        if self._systemSizeWidget is not None:
            self._systemSizeWidget.observe(recomputeFunction, 'value')
        if redrawFunction is not None:
            for widget in self._widgetsPlotOnly.values():
                widget.observe(redrawFunction, 'value')

    def _createAdvancedWidgets(self, _advancedOpts, _continuousReplot: bool = False) -> None:
        """Interface method to add advanced options (if needed)"""
        return None

    def _orderAdvancedWidgets(self, initialState) -> None:
        """Interface method to sort the advanced options, in the self._extraWidgetsOrder list"""
        pass

    def _displayAdvancedOptionsTab(self) -> None:
        """Create and display the "Advanced options" tab (if not empty)"""
        advancedWidgets = []
        atLeastOneAdvancedWidget = False
        for widgetName in self._extraWidgetsOrder:
            if self._widgetsExtraParams.get(widgetName):
                advancedWidgets.append(self._widgetsExtraParams[widgetName])
                if not self._widgetsExtraParams[widgetName].layout.display == 'none':
                    atLeastOneAdvancedWidget = True
            elif self._widgetsPlotOnly.get(widgetName):
                advancedWidgets.append(self._widgetsPlotOnly[widgetName])
                if not self._widgetsPlotOnly[widgetName].layout.display == 'none':
                    atLeastOneAdvancedWidget = True
            # else:
                # print("WARNING! In the _extraWidgetsOrder is listed the widget " + widgetName + " which is although not found in _widgetsExtraParams or _widgetsPlotOnly")
        if advancedWidgets:  # if not empty
            advancedPage = widgets.Box(children=advancedWidgets)
            advancedPage.layout.flex_flow = 'column'
            self._advancedTabWidget = widgets.Accordion(children=[advancedPage])  # , selected_index=-1)
            self._advancedTabWidget.set_title(0, 'Advanced options')
            self._advancedTabWidget.selected_index = None
            if atLeastOneAdvancedWidget:
                display(self._advancedTabWidget)

    def _setView(self, view) -> None:
        self._view = view

    def showLogs(self, tail: bool = False) -> None:
        """Show logs from associated view.

        Parameters
        ----------
        tail : bool, optional
           Flag to show only tail entries from logs. Defaults to False.
        """
        self._view.showLogs(tail)

    def _updateInitialStateWidgets(self, _=None) -> None:
        (allReactants, _) = self._view._mumotModel._getAllReactants()
        if len(allReactants) == 1:
            return
        sumNonConstReactants = 0
        for state in allReactants:
            sumNonConstReactants += self._widgetsExtraParams[f"init{state}"].value
        substitutedReactant = None
        if self._view._mumotModel._systemSize is not None:
            substitutedReactant = [react for react in allReactants
                                   if react not in self._view._mumotModel._reactants][0]
        disabledValue = 1
        for i, state in enumerate(sorted(allReactants, key=str)):
            if (substitutedReactant is None and i == 0) or (substitutedReactant is not None and state == substitutedReactant):
                disabledValue = 1 - (sumNonConstReactants - self._widgetsExtraParams[f"init{state}"].value)
                break

        for i, state in enumerate(sorted(allReactants, key=str)):
            # oder of assignment is important (first, update the min and max, later, the value)
            toLinkPlotFunction = False
            # the self._view._controller pointer is necessary to work properly with multiControllers
            # if self._replotFunction is not None:
            if self._view._controller._replotFunction is not None:
                try:
                    self._widgetsExtraParams[f"init{state}"].unobserve(self._view._controller._replotFunction, 'value')
                    # self._widgetsExtraParams['init'+str(state)].unobserve(self._replotFunction, 'value')
                    toLinkPlotFunction = True
                except ValueError:
                    pass

            disabledState = ((substitutedReactant is None and i == 0) or (substitutedReactant is not None and state == substitutedReactant))
            if disabledState:
                # print(str(state) + ": sum is " + str(sumNonConstReactants) + " - val " + str(disabledValue))
                self._widgetsExtraParams[f"init{state}"].value = disabledValue
            else:
                # maxVal = 1-disabledValue if 1-disabledValue > self._widgetsExtraParams['init'+str(state)].min else self._widgetsExtraParams['init'+str(state)].min
                maxVal = disabledValue + self._widgetsExtraParams[f"init{state}"].value
                self._widgetsExtraParams[f"init{state}"].max = maxVal

            if toLinkPlotFunction:
                self._widgetsExtraParams[f"init{state}"].observe(self._view._controller._replotFunction, 'value')

    def _updateFinalViewWidgets(self, change=None) -> None:
        if change['new'] != 'final':
            if self._widgetsPlotOnly.get('final_x'):
                self._widgetsPlotOnly['final_x'].layout.display = 'none'
            if self._widgetsPlotOnly.get('final_y'):
                self._widgetsPlotOnly['final_y'].layout.display = 'none'
        else:
            if self._widgetsPlotOnly.get('final_x'):
                self._widgetsPlotOnly['final_x'].layout.display = 'flex'
            if self._widgetsPlotOnly.get('final_y'):
                self._widgetsPlotOnly['final_y'].layout.display = 'flex'

    def _setErrorWidget(self, errorWidget) -> None:
        self._errorMessage = errorWidget

    def _downloadFileWithJavascript(self, data_to_download) -> Javascript:
        js_download = """
        var csv = '%s';

        var filename = 'results.csv';
        var blob = new Blob([csv], { type: 'text/csv;charset=utf-8;' });
        if (navigator.msSaveBlob) { // IE 10+
            navigator.msSaveBlob(blob, filename);
        } else {
            var link = document.createElement("a");
            if (link.download !== undefined) { // feature detection
                // Browsers that support HTML5 download attribute
                var url = URL.createObjectURL(blob);
                link.setAttribute("href", url);
                link.setAttribute("download", filename);
                link.style.visibility = 'hidden';
                document.body.appendChild(link);
                link.click();
                document.body.removeChild(link);
            }
        }
        """ % str(data_to_download)
        # str(data_to_download)
        # data_to_download.to_csv(index=False).replace('\n','\\n').replace("'","\'")

        return Javascript(js_download)

    def _create_download_link(self, text: str, title: str = "Download file",
                              filename: str = "file.txt") -> str:
        """Create a download link."""
        b64 = base64.b64encode(text.encode())
        payload = b64.decode()
        html = '<a download="{filename}" href="data:text/text;base64,{payload}" target="_blank">{title}</a>'
        html = html.format(payload=payload, title=title, filename=filename)
        return html

    def _reveal_download_link(self, _includeParams) -> None:
        """Make download link visible."""
        self._downloadWidgetLink.layout.visibility = 'visible'

    def _download_link_unsupported(self, _includeParams) -> None:
        """Report that results download is unsupported."""
        self._view._showErrorMessage("Results download for this view is currently unsupported")


class MuMoTbifurcationController(MuMoTcontroller):
    """Controller to enable Advanced options widgets for bifurcation view."""

    def _createAdvancedWidgets(self, BfcParams, continuousReplot: bool = False):
        initialState = BfcParams['initialState'][0]
        if not BfcParams['initialState'][-1]:
            # for state,pop in initialState.items():
            for i, state in enumerate(sorted(initialState.keys(), key=str)):
                pop = initialState[state]
                widget = widgets.FloatSlider(value=pop[0],
                                             min=pop[1],
                                             max=pop[2],
                                             step=pop[3],
                                             readout_format='.' + str(utils._count_sig_decimals(str(pop[3]))) + 'f',
                                             # r'$' + r'\Phi_{' + utils._doubleUnderscorify(utils._greekPrependify(str(stateVarExpr1))) +'}$'
                                             description=r'$' + r'\Phi_{' + utils._doubleUnderscorify(utils._greekPrependify(self._paramLabelDict.get(state, str(state)))) + '}$' + " at t=0: ",
                                             # description = r'\(' + utils._doubleUnderscorify(utils._greekPrependify('Phi_'+self._paramLabelDict.get(state,str(state)))) + r'\)' + " at t=0: ",
                                             style={'description_width': 'initial'},
                                             continuous_update=continuousReplot)

                if BfcParams['conserved'][0] is True:
                    # Disable one population widget (if there are more than 1) to constrain the initial sum of population sizes to 1
                    if len(initialState) > 1:
                        if BfcParams['substitutedReactant'][0] is None and i == 0:
                            # If there is not a 'substituted' reactant, the last population widget
                            widget.disabled = True
                        elif BfcParams['substitutedReactant'][0] is not None and state == BfcParams['substitutedReactant'][0]:
                            # If there is a 'substituted' reactant, this is the chosen one as the disabled pop
                            widget.disabled = True
                        else:
                            widget.observe(self._updateInitialStateWidgets, 'value')

                self._widgetsExtraParams[f"init{state}"] = widget

        # init bifurcation paramter value slider
        if not BfcParams['initBifParam'][-1]:
            initBifParam = BfcParams['initBifParam']
            widget = widgets.FloatSlider(value=initBifParam[0], min=initBifParam[1],
                                         max=initBifParam[2], step=initBifParam[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(initBifParam[3]))) + 'f',
                                         description='Initial ' + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(str(BfcParams['bifurcationParameter'][0]))) + r'\) : ',
                                         style={'description_width': 'initial:'},
                                         # layout=widgets.Layout(width='50%'),
                                         disabled=False,
                                         continuous_update=continuousReplot)
            self._widgetsExtraParams['initBifParam'] = widget

        return initialState

    def _orderAdvancedWidgets(self, initialState) -> None:
        # define the widget order
        self._extraWidgetsOrder.append('initBifParam')
        for state in sorted(initialState.keys(), key=str):
            self._extraWidgetsOrder.append(f"init{state}")


class MuMoTtimeEvolutionController(MuMoTcontroller):
    """Controller class to enable Advanced options widgets for simulation of ODEs and time evolution of noise correlations."""

    def _createAdvancedWidgets(self, tEParams, continuousReplot=False):
        initialState = tEParams['initialState'][0]
        if not tEParams['initialState'][-1]:
            # for state,pop in initialState.items():
            for i, state in enumerate(sorted(initialState.keys(), key=str)):
                pop = initialState[state]
                widget = widgets.FloatSlider(value=pop[0],
                                             min=pop[1],
                                             max=pop[2],
                                             step=pop[3],
                                             readout_format='.' + str(utils._count_sig_decimals(str(pop[3]))) + 'f',
                                             # description = "Reactant " + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(self._paramLabelDict.get(state,str(state)))) + r'\)' + " at t=0: ",
                                             description=r'$' + r'\Phi_{' + utils._doubleUnderscorify(utils._greekPrependify(self._paramLabelDict.get(state, str(state)))) + '}$' + " at t=0: ",
                                             # description = r'\(' + latex(Symbol('Phi_'+str(state))) + r'\)' + " at t=0: ",
                                             style={'description_width': 'initial'},
                                             continuous_update=continuousReplot)

                if tEParams['conserved'][0] is True:
                    # Disable one population widget (if there are more than 1) to constrain the initial sum of population sizes to 1
                    if len(initialState) > 1:
                        if tEParams['substitutedReactant'][0] is None and i == 0:
                            # If there is not a 'substituted' reactant, the last population widget
                            widget.disabled = True
                        elif tEParams['substitutedReactant'][0] is not None and state == tEParams['substitutedReactant'][0]:
                            # If there is a 'substituted' reactant, this is the chosen one as the disabled pop
                            widget.disabled = True
                        else:
                            widget.observe(self._updateInitialStateWidgets, 'value')

                self._widgetsExtraParams[f"init{state}"] = widget

        # Max time slider
        if not tEParams['maxTime'][-1]:
            maxTime = tEParams['maxTime']
            widget = widgets.FloatSlider(value=maxTime[0], min=maxTime[1],
                                         max=maxTime[2], step=maxTime[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(maxTime[3]))) + 'f',
                                         description='Simulation time:',
                                         style={'description_width': 'initial'},
                                         # layout=widgets.Layout(width='50%'),
                                         disabled=False,
                                         continuous_update=continuousReplot)
            self._widgetsExtraParams['maxTime'] = widget

        # Checkbox for proportions or full populations plot
        if 'plotProportions' in tEParams:
            if not tEParams['plotProportions'][-1]:
                widget = widgets.Checkbox(
                    value=tEParams['plotProportions'][0],
                    description='Plot population proportions',
                    disabled=False
                )
                self._widgetsPlotOnly['plotProportions'] = widget

        return initialState

    def _orderAdvancedWidgets(self, initialState):
        # define the widget order
        for state in sorted(initialState.keys(), key=str):
            self._extraWidgetsOrder.append(f"init{state}")
        self._extraWidgetsOrder.append('maxTime')
        if 'plotProportions' in self._widgetsPlotOnly:
            self._extraWidgetsOrder.append('plotProportions')


class MuMoTfieldController(MuMoTcontroller):
    """Controller for field views"""

    def _createAdvancedWidgets(self, advancedOpts, continuousReplot=False):
        # Max time slider
        if not advancedOpts['maxTime'][-1]:
            maxTime = advancedOpts['maxTime']
            widget = widgets.FloatSlider(value=maxTime[0], min=maxTime[1],
                                         max=maxTime[2], step=maxTime[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(maxTime[3]))) + 'f',
                                         description='Simulation time:',
                                         style={'description_width': 'initial'},
                                         # layout=widgets.Layout(width='50%'),
                                         disabled=False,
                                         continuous_update=continuousReplot)
            self._widgetsExtraParams['maxTime'] = widget

        # Random seed input field
        if not advancedOpts['randomSeed'][-1]:
            widget = widgets.IntText(
                value=advancedOpts['randomSeed'][0],
                description='Random seed:',
                style={'description_width': 'initial'},
                disabled=False
            )
            self._widgetsExtraParams['randomSeed'] = widget

        # @todo: this block of commented code can be used readily used to fix issue #95
        # if not advancedOpts['final_x'][-1]:
        #     opts = []
        #     for reactant in sorted(initialState.keys(), key=str):
        #         # opts.append( ( "Reactant " + r'$'+ latex(Symbol(str(reactant)))+r'$', str(reactant) ) )
        #         opts.append(("Reactant " + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) + r'\)', str(reactant)))
        #     dropdown = widgets.Dropdown(
        #         options=opts,
        #         description='Final distribution (x axis):',
        #         value=advancedOpts['final_x'][0],
        #         style={'description_width': 'initial'}
        #     )
        #     self._widgetsPlotOnly['final_x'] = dropdown
        # if not advancedOpts['final_y'][-1]:
        #     opts = []
        #     for reactant in sorted(initialState.keys(), key=str):
        #           opts.append( ( r'$'+ utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) +'$' , str(reactant) ) )
        #           print("the reactant is " + str(reactant))
        #           print("the utils._greekPrependify(str(reactant) is " + str(utils._greekPrependify(str(reactant))) )
        #           print("the utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) is " + str(utils._doubleUnderscorify(utils._greekPrependify(str(reactant)))) )
        #         opts.append(("Reactant " + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) + r'\)', str(reactant)))
        #     dropdown = widgets.Dropdown(
        #         options=opts,
        #         description='Final distribution (y axis):',
        #         value=advancedOpts['final_y'][0],
        #         style={'description_width': 'initial'}
        #     )
        #     self._widgetsPlotOnly['final_y'] = dropdown

        # @todo: this block of commented code can be used readily used to fix issue #283
        # # Checkbox for proportions or full populations plot
        # if not advancedOpts['plotProportions'][-1]:
        #     widget = widgets.Checkbox(
        #         value=advancedOpts['plotProportions'][0],
        #         description='Plot population proportions',
        #         disabled=False
        #     )
        #     self._widgetsPlotOnly['plotProportions'] = widget

        # Number of runs slider
        if not advancedOpts['runs'][-1]:
            runs = advancedOpts['runs']
            widget = widgets.IntSlider(value=runs[0], min=runs[1],
                                       max=runs[2], step=runs[3],
                                       readout_format='.' + str(utils._count_sig_decimals(str(runs[3]))) + 'f',
                                       description='Number of runs:',
                                       style={'description_width': 'initial'},
                                       # layout=widgets.Layout(width='50%'),
                                       disabled=False,
                                       continuous_update=continuousReplot)
            self._widgetsExtraParams['runs'] = widget

        # Checkbox for realtime plot update
        if not advancedOpts['aggregateResults'][-1]:
            widget = widgets.Checkbox(
                value=advancedOpts['aggregateResults'][0],
                description='Aggregate results',
                disabled=False
            )
            self._widgetsExtraParams['aggregateResults'] = widget

        return None

    def _orderAdvancedWidgets(self, _noInitialState):
        # define the widget order
        # self._extraWidgetsOrder.append('final_x')
        # self._extraWidgetsOrder.append('final_y')
        # self._extraWidgetsOrder.append('plotProportions')
        self._extraWidgetsOrder.append('runs')
        self._extraWidgetsOrder.append('maxTime')
        self._extraWidgetsOrder.append('randomSeed')
        self._extraWidgetsOrder.append('aggregateResults')


class MuMoTstochasticSimulationController(MuMoTcontroller):
    """Controller for stochastic simulations (base class of MuMoTmultiagentController)."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._downloadWidget.on_click(self._download_link_unsupported, remove=True)
        self._downloadWidget.on_click(self._reveal_download_link)

    def _createAdvancedWidgets(self, SSParams, continuousReplot=False):
        initialState = SSParams['initialState'][0]
        if not SSParams['initialState'][-1]:
            # for state,pop in initialState.items():
            for i, state in enumerate(sorted(initialState.keys(), key=str)):
                pop = initialState[state]
                widget = widgets.FloatSlider(value=pop[0],
                                             min=pop[1],
                                             max=pop[2],
                                             step=pop[3],
                                             readout_format='.' + str(utils._count_sig_decimals(str(pop[3]))) + 'f',
                                             description=r'$' + utils._doubleUnderscorify(utils._greekPrependify(str(Symbol('Phi_{' + str(state) + '}')))) + '$' + " at t=0: ",
                                             # description = r'\(' + latex(Symbol('Phi_'+str(state))) + r'\)' + " at t=0: ",
                                             style={'description_width': 'initial'},
                                             continuous_update=continuousReplot)
                # disable one population widget (if there are more than 1) to constrain the initial sum of population sizes to 1
                if len(initialState) > 1:
                    if SSParams['substitutedReactant'][0] is None and i == 0:
                        # if there is not a 'substituted' reactant, the last population widget
                        widget.disabled = True
                    elif SSParams['substitutedReactant'][0] is not None and state == SSParams['substitutedReactant'][0]:
                        # if there is a 'substituted' reactant, this is the chosen one as the disabled pop
                        widget.disabled = True
                    else:
                        widget.observe(self._updateInitialStateWidgets, 'value')

                self._widgetsExtraParams[f"init{state}"] = widget

        # Max time slider
        if not SSParams['maxTime'][-1]:
            maxTime = SSParams['maxTime']
            widget = widgets.FloatSlider(value=maxTime[0], min=maxTime[1],
                                         max=maxTime[2], step=maxTime[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(maxTime[3]))) + 'f',
                                         description='Simulation time:',
                                         style={'description_width': 'initial'},
                                         # layout=widgets.Layout(width='50%'),
                                         disabled=False,
                                         continuous_update=continuousReplot)
            self._widgetsExtraParams['maxTime'] = widget

        # Random seed input field
        if not SSParams['randomSeed'][-1]:
            widget = widgets.IntText(
                value=SSParams['randomSeed'][0],
                description='Random seed:',
                style={'description_width': 'initial'},
                disabled=False
            )
            self._widgetsExtraParams['randomSeed'] = widget

        try:
            # Toggle buttons for plotting style
            if not SSParams['visualisationType'][-1]:
                plotToggle = widgets.ToggleButtons(
                    options=[('Temporal evolution', 'evo'), ('Final distribution', 'final'), ('Barplot', 'barplot')],
                    value=SSParams['visualisationType'][0],
                    description='Plot:',
                    disabled=False,
                    button_style='',  # 'success', 'info', 'warning', 'danger' or ''
                    tooltips=['Population change over time', 'Population distribution in each state at final timestep', 'Barplot of states at final timestep'],
                    # icons=['check'] * 3
                )
                plotToggle.observe(self._updateFinalViewWidgets, 'value')
                self._widgetsPlotOnly['visualisationType'] = plotToggle

        except widgets.trait_types.traitlets.TraitError:  # this widget could be redefined in a subclass and the init-value in SSParams['visualisationType'][0] might raise an exception
            pass

        if not SSParams['final_x'][-1] and (SSParams['visualisationType'][-1] is False or SSParams['visualisationType'][0] == 'final'):
            opts = []
            for reactant in sorted(initialState.keys(), key=str):
                # opts.append( ( "Reactant " + r'$'+ latex(Symbol(str(reactant)))+r'$', str(reactant) ) )
                opts.append(("Reactant " + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) + r'\)', str(reactant)))
            dropdown = widgets.Dropdown(
                options=opts,
                description='Final distribution (x axis):',
                value=SSParams['final_x'][0],
                style={'description_width': 'initial'}
            )
            if SSParams['visualisationType'][0] != 'final':
                dropdown.layout.display = 'none'
            self._widgetsPlotOnly['final_x'] = dropdown
        if not SSParams['final_y'][-1] and (SSParams['visualisationType'][-1] is False or SSParams['visualisationType'][0] == 'final'):
            opts = []
            for reactant in sorted(initialState.keys(), key=str):
                # opts.append( ( r'$'+ utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) +'$' , str(reactant) ) )
                # print("the reactant is " + str(reactant))
                # print("the utils._greekPrependify(str(reactant) is " + str(utils._greekPrependify(str(reactant))) )
                # print("the utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) is " + str(utils._doubleUnderscorify(utils._greekPrependify(str(reactant)))) )
                opts.append(("Reactant " + r'\(' + utils._doubleUnderscorify(utils._greekPrependify(str(reactant))) + r'\)', str(reactant)))
            dropdown = widgets.Dropdown(
                options=opts,
                description='Final distribution (y axis):',
                value=SSParams['final_y'][0],
                style={'description_width': 'initial'}
            )
            if SSParams['visualisationType'][0] != 'final':
                dropdown.layout.display = 'none'
            self._widgetsPlotOnly['final_y'] = dropdown

        # Checkbox for proportions or full populations plot
        if not SSParams['plotProportions'][-1]:
            widget = widgets.Checkbox(
                value=SSParams['plotProportions'][0],
                description='Plot population proportions',
                disabled=False
            )
            self._widgetsPlotOnly['plotProportions'] = widget

        # Checkbox for realtime plot update
        if not SSParams['realtimePlot'][-1]:
            widget = widgets.Checkbox(
                value=SSParams['realtimePlot'][0],
                description='Runtime plot update',
                disabled=False
            )
            self._widgetsExtraParams['realtimePlot'] = widget

        # Number of runs slider
        if not SSParams['runs'][-1]:
            runs = SSParams['runs']
            widget = widgets.IntSlider(value=runs[0], min=runs[1],
                                       max=runs[2], step=runs[3],
                                       readout_format='.' + str(utils._count_sig_decimals(str(runs[3]))) + 'f',
                                       description='Number of runs:',
                                       style={'description_width': 'initial'},
                                       # layout=widgets.Layout(width='50%'),
                                       disabled=False,
                                       continuous_update=continuousReplot)
            self._widgetsExtraParams['runs'] = widget

        # Checkbox for realtime plot update
        if not SSParams['aggregateResults'][-1]:
            widget = widgets.Checkbox(
                value=SSParams['aggregateResults'][0],
                description='Aggregate results',
                disabled=False
            )
            self._widgetsPlotOnly['aggregateResults'] = widget

        self._addSpecificWidgets(SSParams, continuousReplot)

        return initialState

    def _addSpecificWidgets(self, SSParams, continuousReplot):
        pass

    def _orderAdvancedWidgets(self, initialState):
        # define the widget order
        for state in sorted(initialState.keys(), key=str):
            self._extraWidgetsOrder.append(f"init{state}")
        self._extraWidgetsOrder.append('maxTime')
        self._extraWidgetsOrder.append('randomSeed')
        self._extraWidgetsOrder.append('visualisationType')
        self._extraWidgetsOrder.append('final_x')
        self._extraWidgetsOrder.append('final_y')
        self._extraWidgetsOrder.append('plotProportions')
        self._extraWidgetsOrder.append('realtimePlot')
        self._extraWidgetsOrder.append('runs')
        self._extraWidgetsOrder.append('aggregateResults')


class MuMoTmultiagentController(MuMoTstochasticSimulationController):
    """Controller for multiagent views."""

    def _addSpecificWidgets(self, MAParams, continuousReplot=False):

        # Network type dropdown selector
        if not MAParams['netType'][-1]:
            netDropdown = widgets.Dropdown(
                options=[('Full graph', consts.NetworkType.FULLY_CONNECTED),
                         ('Erdos-Renyi', consts.NetworkType.ERSOS_RENYI),
                         ('Barabasi-Albert', consts.NetworkType.BARABASI_ALBERT),
                         # @todo: add network topology generated by random points in space
                         ('Moving particles', consts.NetworkType.DYNAMIC)
                         ],
                description='Network topology:',
                value=utils._decodeNetworkTypeFromString(MAParams['netType'][0]),
                style={'description_width': 'initial'},
                disabled=False
            )
            netDropdown.observe(self._update_net_params, 'value')
            self._widgetsExtraParams['netType'] = netDropdown

        # Network connectivity slider
        if not MAParams['netParam'][-1]:
            netParam = MAParams['netParam']
            widget = widgets.FloatSlider(value=netParam[0],
                                         min=netParam[1],
                                         max=netParam[2],
                                         step=netParam[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(netParam[3]))) + 'f',
                                         description='Network connectivity parameter',
                                         style={'description_width': 'initial'},
                                         layout=widgets.Layout(width='50%'),
                                         continuous_update=continuousReplot,
                                         disabled=False)
            self._widgetsExtraParams['netParam'] = widget

        # Agent speed
        if not MAParams['particleSpeed'][-1]:
            particleSpeed = MAParams['particleSpeed']
            widget = widgets.FloatSlider(value=particleSpeed[0],
                                         min=particleSpeed[1], max=particleSpeed[2], step=particleSpeed[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(particleSpeed[3]))) + 'f',
                                         description='Particle speed',
                                         style={'description_width': 'initial'},
                                         layout=widgets.Layout(width='50%'),
                                         continuous_update=continuousReplot,
                                         disabled=False)
            self._widgetsExtraParams['particleSpeed'] = widget

        # Random walk correlatedness
        if not MAParams['motionCorrelatedness'][-1]:
            motionCorrelatedness = MAParams['motionCorrelatedness']
            widget = widgets.FloatSlider(value=motionCorrelatedness[0],
                                         min=motionCorrelatedness[1],
                                         max=motionCorrelatedness[2],
                                         step=motionCorrelatedness[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(motionCorrelatedness[3]))) + 'f',
                                         description='Correlatedness of the random walk',
                                         layout=widgets.Layout(width='50%'),
                                         style={'description_width': 'initial'},
                                         continuous_update=continuousReplot,
                                         disabled=False)
            self._widgetsExtraParams['motionCorrelatedness'] = widget

        # Time scaling slider
        if not MAParams['timestepSize'][-1]:
            timestepSize = MAParams['timestepSize']
            widget = widgets.FloatSlider(value=timestepSize[0],
                                         min=timestepSize[1],
                                         max=timestepSize[2],
                                         step=timestepSize[3],
                                         readout_format='.' + str(utils._count_sig_decimals(str(timestepSize[3]))) + 'f',
                                         description='Timestep size',
                                         style={'description_width': 'initial'},
                                         layout=widgets.Layout(width='50%'),
                                         continuous_update=continuousReplot)
            self._widgetsExtraParams['timestepSize'] = widget

        # Toggle buttons for plotting style
        if not MAParams['visualisationType'][-1]:
            plotToggle = widgets.ToggleButtons(
                options=[('Temporal evolution', 'evo'), ('Network', 'graph'), ('Final distribution', 'final'), ('Barplot', 'barplot')],
                value=MAParams['visualisationType'][0],
                description='Plot:',
                disabled=False,
                button_style='',  # 'success', 'info', 'warning', 'danger' or ''
                tooltips=['Population change over time', 'Population distribution in each state at final timestep', 'Barplot of states at final timestep'],
            )
            plotToggle.observe(self._updateFinalViewWidgets, 'value')
            self._widgetsPlotOnly['visualisationType'] = plotToggle

        # Particle display checkboxes
        if not MAParams['showTrace'][-1]:
            widget = widgets.Checkbox(
                value=MAParams['showTrace'][0],
                description='Show particle trace',
                disabled=False  # not (self._widgetsExtraParams['netType'].value == consts.NetworkType.DYNAMIC)
            )
            self._widgetsPlotOnly['showTrace'] = widget
        if not MAParams['showInteractions'][-1]:
            widget = widgets.Checkbox(
                value=MAParams['showInteractions'][0],
                description='Show communication links',
                disabled=False  # not (self._widgetsExtraParams['netType'].value == consts.NetworkType.DYNAMIC)
            )
            self._widgetsPlotOnly['showInteractions'] = widget

    def _orderAdvancedWidgets(self, initialState):
        # define the widget order
        for state in sorted(initialState.keys(), key=str):
            self._extraWidgetsOrder.append(f"init{state}")
        self._extraWidgetsOrder.append('maxTime')
        self._extraWidgetsOrder.append('timestepSize')
        self._extraWidgetsOrder.append('netType')
        self._extraWidgetsOrder.append('netParam')
        self._extraWidgetsOrder.append('particleSpeed')
        self._extraWidgetsOrder.append('motionCorrelatedness')
        self._extraWidgetsOrder.append('randomSeed')
        self._extraWidgetsOrder.append('visualisationType')
        self._extraWidgetsOrder.append('final_x')
        self._extraWidgetsOrder.append('final_y')
        self._extraWidgetsOrder.append('showTrace')
        self._extraWidgetsOrder.append('showInteractions')
        self._extraWidgetsOrder.append('plotProportions')
        self._extraWidgetsOrder.append('realtimePlot')
        self._extraWidgetsOrder.append('runs')
        self._extraWidgetsOrder.append('aggregateResults')

    def _update_net_params(self, _=None):
        """Update the widgets related to the ``netType``

        It is linked - through ``observe()`` - before the ``_view`` is created.

        """
        if self._view:
            self._view._update_net_params(True)


class MuMoTmultiController(MuMoTcontroller):
    """Multi-view controller."""

    # replot function list to invoke on views
    _replotFunctions = None

    def __init__(self, controllers, params=None, initWidgets=None, **kwargs):
        if initWidgets is None:
            initWidgets = {}

        self._silent = kwargs.get('silent', False)
        self._replotFunctions = []
        fixedParamNames = None
        paramValuesDict = {}
        paramLabelDict = {}
        showPlotLimits = False
        showSystemSize = False
        views_ = []
        subPlotNum = 1
        model = None
        # @todo assuming same model for all views.
        # This operation is NOT correct when multicotroller views have different models
        # paramValuesDict = controllers[0]._view._mumotModel._create_free_param_dictionary_for_controller(inputParams=params if params is not None else [], initWidgets=initWidgets, showSystemSize=True, showPlotLimits=True )

        if params is not None:
            (fixedParamNames, fixedParamValues) = utils._process_params(params)
        for controller in controllers:
            # pass through the fixed params to each constituent view
            view = controller._view
            if params is not None:
                # view._fixedParams = dict(zip(fixedParamNames, fixedParamValues))
                # view._fixedParams = {**dict(zip(fixedParamNames, fixedParamValues)), **view._fixedParams}
                #     this operation merge the two dictionaries with the second overriding the values of the first

                # This operation merges the two dictionaries with the second overriding the values of the first.
                view._set_fixedParams({**dict(zip(fixedParamNames, fixedParamValues)), **view._fixedParams})
            for name, value in controller._widgetsFreeParams.items():
                # if params is None or name not in fixedParamNames:
                #    paramValueDict[name] = (value.value, value.min, value.max, value.step)
                if name in initWidgets:
                    paramValuesDict[name] = utils._parse_input_keyword_for_numeric_widgets(
                        inputValue=utils._get_item_from_params_list(params if params is not None else [], name),
                        defaultValueRangeStep=[defaults.MuMoTdefault._initialRateValue,
                                               defaults.MuMoTdefault._rateLimits[0],
                                               defaults.MuMoTdefault._rateLimits[1],
                                               defaults.MuMoTdefault._rateStep],
                        initValueRangeStep=initWidgets.get(name),
                        validRange=(-float("inf"), float("inf")))
                else:
                    paramValuesDict[name] = (value.value,
                                             value.min,
                                             value.max,
                                             value.step,
                                             not(params is None or name not in map(str, view._fixedParams.keys())))
            if controller._plotLimitsWidget is not None:
                showPlotLimits = True
                if 'plotLimits' in initWidgets:
                    paramValuesDict['plotLimits'] = utils._parse_input_keyword_for_numeric_widgets(
                        inputValue=utils._get_item_from_params_list(params if params is not None else [], 'plotLimits'),
                        defaultValueRangeStep=[defaults.MuMoTdefault._plotLimits,
                                               defaults.MuMoTdefault._plotLimitsLimits[0],
                                               defaults.MuMoTdefault._plotLimitsLimits[1],
                                               defaults.MuMoTdefault._plotLimitsStep],
                        initValueRangeStep=initWidgets.get('plotLimits'),
                        validRange=(-float("inf"), float("inf")))
                else:
                    paramValuesDict['plotLimits'] = (controller._plotLimitsWidget.value,
                                                     controller._plotLimitsWidget.min,
                                                     controller._plotLimitsWidget.max,
                                                     controller._plotLimitsWidget.step,
                                                     not(params is None or 'plotLimits' not in map(str, view._fixedParams.keys())))
            if controller._systemSizeWidget is not None:
                showSystemSize = True
                if 'systemSize' in initWidgets:
                    paramValuesDict['systemSize'] = utils._parse_input_keyword_for_numeric_widgets(
                        inputValue=utils._get_item_from_params_list(params if params is not None else [], 'systemSize'),
                        defaultValueRangeStep=[defaults.MuMoTdefault._systemSize,
                                               defaults.MuMoTdefault._systemSizeLimits[0],
                                               defaults.MuMoTdefault._systemSizeLimits[1],
                                               defaults.MuMoTdefault._systemSizeStep],
                        initValueRangeStep=initWidgets.get('systemSize'),
                        validRange=(1, float("inf")))
                else:
                    paramValuesDict['systemSize'] = (controller._systemSizeWidget.value,
                                                     controller._systemSizeWidget.min,
                                                     controller._systemSizeWidget.max,
                                                     controller._systemSizeWidget.step,
                                                     not(params is None or 'systemSize' not in map(str, view._fixedParams.keys())))
            paramLabelDict.update(controller._paramLabelDict)
            # for name, value in controller._widgetsExtraParams.items():
            # widgetsExtraParamsTmp[name] = value
            if type(controller) is MuMoTmultiController:
                # if controller._replotFunction is None: # presume this controller is a multi controller (@todo check?)
                for view in controller._view._views:
                    views_.append(view)

                # if view._controller._replotFunction is None: # presume this controller is a multi controller (@todo check?)
                for func, _, axes3d in controller._replotFunctions:
                    self._replotFunctions.append((func, subPlotNum, axes3d))
                # else:
                #    self._replotFunctions.append((view._controller._replotFunction, subPlotNum, view._axes3d))
            else:
                views_.append(controller._view)
                # if controller._replotFunction is None: # presume this controller is a multi controller (@todo check?)
                #     for func, foo in controller._replotFunctions:
                #         self._replotFunctions.append((func, subPlotNum))
                # else:
                self._replotFunctions.append((controller._replotFunction,
                                              subPlotNum,
                                              controller._view._axes3d))
            subPlotNum += 1
            # check if all views refer to same model
            if model is None:
                model = view._mumotModel
            elif model != view._mumotModel:
                raise exceptions.MuMoTValueError(
                    'Multicontroller views do not all refer to same model')

        # for view in self._views:
        #     # presume this controller is a multi controller (@todo check?)
        #     if view._controller._replotFunction is None:
        #         for func in view._controller._replotFunctions:
        #             self._replotFunctions.append(func)
        #     else:
        #         self._replotFunctions.append(view._controller._replotFunction)
        #     view._controller = self
        super().__init__(paramValuesDict, paramLabelDict, False,
                         showPlotLimits, showSystemSize, params=params,
                         **kwargs)

        # handle Extra and PlotOnly params
        addProgressBar = False
        for controller in controllers:
            # retrieve the _widgetsExtraParams from each controller
            for name, widget in controller._widgetsExtraParams.items():
                widget.unobserve(controller._replotFunction, 'value')
                self._widgetsExtraParams[name] = widget
            # retrieve the _widgetsPlotOnly from each controller
            for name, widget in controller._widgetsPlotOnly.items():
                widget.unobserve(controller._redrawFunction, 'value')
                # in multiController, plotProportions=True is forced
                if name == "plotProportions":
                    widget.value = True
                    widget.disabled = True
                # this is necessary due to limit visualisation-type to only 'evo' and 'final'
                if name == "visualisationType":
                    # Toggle buttons for plotting style
                    widget = widgets.ToggleButtons(
                        options=[('Temporal evolution', 'evo'), ('Final distribution', 'final')],
                        value=widget.value,
                        description='Plot:',
                        disabled=False,
                        button_style='',  # 'success', 'info', 'warning', 'danger' or ''
                        tooltips=['Population change over time', 'Population distribution in each state at final timestep'],
                    )
                    widget.observe(self._updateFinalViewWidgets, 'value')
                self._widgetsPlotOnly[name] = widget
            # retrieve the _extraWidgetsOrder from each controller
            self._extraWidgetsOrder.extend(x for x in controller._extraWidgetsOrder
                                           if x not in self._extraWidgetsOrder)
            # if controller._progressBar:
            #     addProgressBar = True
        if self._widgetsExtraParams or self._widgetsPlotOnly:
            # set widgets to possible initial/fixed values if specified in the multi-controller
            # for key, value in kwargs.items():
            for key in kwargs.keys() | initWidgets.keys():
                inputValue = kwargs.get(key)
                ep1 = None
                ep2 = None
                if key == 'choose_yrange':
                    for controller in controllers:
                        controller._view._chooseYrange = kwargs.get('choose_yrange')
                if key == 'choose_xrange':
                    for controller in controllers:
                        controller._view._chooseXrange = kwargs.get('choose_xrange')
                if key == 'initialState':
                    ep1 = views_[0]._mumotModel._getAllReactants()
                    ep2 = [True, [react for react in views_[0]._mumotModel._getAllReactants()[0] if react not in views_[0]._mumotModel._reactants][0] if views_[0]._mumotModel._systemSize is not None else None]
                    # @todo assuming same model for all views.
                    # This operation is NOT correct when multicotroller views have different models.
                if key == 'visualisationType':
                    ep1 = "multicontroller"
                if key == 'final_x' or key == 'final_y':
                    ep1 = views_[0]._mumotModel._getAllReactants()[0]
                    # @todo assuming same model for all views.
                    # This operation is NOT correct when multicotroller views have different models.
                if key == 'netParam':
                    ep1 = [kwargs.get('netType', self._widgetsExtraParams.get('netType')),
                           kwargs.get('netType') is not None]
                    maxSysSize = 1
                    for view in views_:
                        maxSysSize = max(maxSysSize, view._getSystemSize())
                    ep2 = maxSysSize
                optionValues = utils._format_advanced_option(
                    optionName=key,
                    inputValue=inputValue,
                    initValues=initWidgets.get(key),
                    extraParam=ep1,
                    extraParam2=ep2)
                # if option is fixed
                if optionValues[-1] is True:
                    if key == 'initialState':  # initialState is special
                        for state, pop in optionValues[0].items():
                            optionValues[0][state] = pop[0]
                            stateKey = "init" + str(state)
                            # delete the widgets
                            if stateKey in self._widgetsExtraParams:
                                del self._widgetsExtraParams[stateKey]
                    if key == 'netType':  # netType is special
                        optionValues[0] = utils._decodeNetworkTypeFromString(optionValues[0])  # @todo: if only netType (and not netParam) is specified, then multicotroller won't work...
                    if key == 'visualisationType' and optionValues[0] == 'final':  # visualisationType == 'final' is special
                        if self._widgetsPlotOnly.get('final_x') is not None:
                            self._widgetsPlotOnly['final_x'].layout.display = 'flex'
                        if self._widgetsPlotOnly.get('final_y') is not None:
                            self._widgetsPlotOnly['final_y'].layout.display = 'flex'
                    # set the value in all the views
                    for view in views_:
                        view._fixedParams[key] = optionValues[0]
                    # delete the widgets
                    if key in self._widgetsExtraParams:
                        del self._widgetsExtraParams[key]
                    if key in self._widgetsPlotOnly:
                        del self._widgetsPlotOnly[key]
                else:
                    # update the values with the init values
                    if key in self._widgetsExtraParams:
                        if len(optionValues) == 5:
                            self._widgetsExtraParams[key].max = 10**7  # temp to avoid exception min>max
                            self._widgetsExtraParams[key].min = optionValues[1]
                            self._widgetsExtraParams[key].max = optionValues[2]
                            self._widgetsExtraParams[key].step = optionValues[3]
                            self._widgetsExtraParams[key].readout_format = '.' + str(utils._count_sig_decimals(str(optionValues[3]))) + 'f'
                        self._widgetsExtraParams[key].value = optionValues[0]
                    if key in self._widgetsPlotOnly:
                        if len(optionValues) == 5:
                            self._widgetsPlotOnly[key].max = 10**7  # temp to avoid exception min>max
                            self._widgetsPlotOnly[key].min = optionValues[1]
                            self._widgetsPlotOnly[key].max = optionValues[2]
                            self._widgetsPlotOnly[key].step = optionValues[3]
                            self._widgetsPlotOnly[key].readout_format = '.' + str(utils._count_sig_decimals(str(optionValues[3]))) + 'f'
                        self._widgetsPlotOnly[key].value = optionValues[0]
                    if key == 'initialState':
                        for state, pop in optionValues[0].items():
                            # self._widgetsExtraParams['init'+str(state)].unobserve(self._updateInitialStateWidgets, 'value')
                            self._widgetsExtraParams[f"init{state}"].max = float('inf')  # temp to avoid exception min>max
                            self._widgetsExtraParams[f"init{state}"].min = pop[1]
                            self._widgetsExtraParams[f"init{state}"].max = pop[2]
                            self._widgetsExtraParams[f"init{state}"].step = pop[3]
                            self._widgetsExtraParams[f"init{state}"].readout_format = '.' + str(utils._count_sig_decimals(str(pop[3]))) + 'f'
                            self._widgetsExtraParams[f"init{state}"].value = pop[0]
                            # self._widgetsExtraParams['init'+str(state)].observe(self._updateInitialStateWidgets, 'value')

            # create the "Advanced options" tab
            if not self._silent:
                self._displayAdvancedOptionsTab()

        # if necessary adding the progress bar
        if addProgressBar:
            # Loading bar (useful to give user progress status for long executions)
            self._progressBar = widgets.FloatProgress(
                value=0,
                min=0,
                max=(self._widgetsExtraParams['maxTime'].value
                     if self._widgetsExtraParams.get('maxTime') is not None
                     else views_[0]._fixedParams.get('maxTime')),
                # step=1,
                description='Loading:',
                bar_style='success',  # 'success', 'info', 'warning', 'danger' or ''
                style={'description_width': 'initial'},
                orientation='horizontal'
            )
            if not self._silent:
                display(self._progressBar)

        self._view = views.MuMoTmultiView(self, model, views_, controllers, subPlotNum - 1, **kwargs)
        if fixedParamNames is not None:
            # self._view._fixedParams = dict(zip(fixedParamNames, fixedParamValues))
            self._view._set_fixedParams(dict(zip(fixedParamNames, fixedParamValues)))

        for controller in controllers:
            controller._setErrorWidget(self._errorMessage)

        # @todo handle correctly the re-draw only widgets and function
        self._setReplotFunction(self._view._plot, self._view._plot)

        # silent = kwargs.get('silent', False)
        if not self._silent:
            self._view._plot()



"""MuMoT warning, exception and error classes."""

class MuMoTWarning(Warning):
    """Class to report MuMoT-specific warnings.
    """
    pass


class MuMoTError(Exception):
    """Class to report MuMoT-specific errors.
    """
    pass


class MuMoTValueError(MuMoTError):
    """Class to report MuMoT-specific errors arising from incorrect input.
    """
    pass


class MuMoTSyntaxError(MuMoTError):
    """Class to report MuMoT-specific errors arising from incorrectly-structured input.
    """
    pass


def _raiseModelError(expected, read, rule):
    raise MuMoTSyntaxError(f"Expected {expected} but read '{read}' in rule: {rule}")



"""Multiscale Modelling Tool (MuMoT)

For documentation and version information use about()

Authors:
James A. R. Marshall, Andreagiovanni Reina, Thomas Bose

Contributors:
Robert Dennison

Packaging, Documentation and Deployment:
Will Furnass

Windows Compatibility:
Renato Pagliara Vasquez
"""

import sys

import matplotlib
# If operating system is macOS use a non-default matplotlib backend
# otherwise rendering of images might not be correct
# (e.g. superfluous figures when sliders are moved).
# Automated testing using tox could be affected as well
# if default matplotlib backend is used
#if sys.platform == "darwin":
matplotlib.use('TkAgg')
from matplotlib import pyplot as plt
# Guard against iopub rate limiting warnings (https://github.com/DiODeProject/MuMoT/issues/359)
from notebook.notebookapp import NotebookApp
NotebookApp.iopub_msg_rate_limit = 10000.0
from sympy.parsing.latex import parse_latex

# Import the functions and classes we wish to export i.e. the public API
from ._version import __version__
from .models import (
    MuMoTmodel,
    parseModel,
)
from .utils import (
    about,
)
from .views import (
    MuMoTSSAView,
    MuMoTbifurcationView,
    MuMoTfieldView,
    MuMoTintegrateView,
    MuMoTmultiView,
    MuMoTmultiagentView,
    MuMoTnoiseCorrelationsView,
    MuMoTstochasticSimulationView,
    MuMoTstreamView,
    MuMoTtimeEvolutionView,
    MuMoTvectorView,
    MuMoTview,
)
from .controllers import (
    MuMoTbifurcationController,
    MuMoTcontroller,
    MuMoTfieldController,
    MuMoTmultiController,
    MuMoTmultiagentController,
    MuMoTstochasticSimulationController,
    MuMoTtimeEvolutionController,
)
from .consts import (
    NetworkType,
    MAX_RANDOM_SEED,
)
from .exceptions import (
    MuMoTError,
    MuMoTSyntaxError,
    MuMoTValueError,
    MuMoTWarning,
)

try:
    # Try to get the currently-running IPython instance
    ipython = get_ipython()
    ipython.magic('alias_magic model latex')
    ipython.magic('matplotlib nbagg')

    def _hide_traceback(exc_tuple=None, filename=None, tb_offset=None,
                        exception_only=False, running_compiled_code=False):
        etype, value, tb = sys.exc_info()
        return ipython._showtraceback(etype, value, ipython.InteractiveTB.get_exception_only(etype, value))

    _show_traceback = ipython.showtraceback
    ipython.showtraceback = _hide_traceback
except NameError:
    # There is no currently-running IPython instance
    pass


def setVerboseExceptions(verbose: bool = True) -> None:
    """Set the verbosity of exception handling.

    Parameters
    ----------
    verbose : bool, optional
        Whether to show a exception traceback.  Defaults to True.

    """
    ipython.showtraceback = _show_traceback if verbose else _hide_traceback



import math
import numbers
from typing import Optional, List

import numpy as np
from sympy.parsing.latex import parse_latex
from warnings import warn

from . import (
    consts,
    defaults,
    exceptions,
    utils,
    _version
)


def about() -> None:
    """Print version, author and documentation information."""
    print("Multiscale Modelling Tool (MuMoT): Version " + _version.__version__)
    print("Authors: James A. R. Marshall, Andreagiovanni Reina, Thomas Bose")
    print("Contributors: Robert Dennison, Will Furnass")
    print("Documentation: https://mumot.readthedocs.io/")


def _greekPrependify(s: str) -> str:
    """Prepend two backslash symbols in front of Greek letters to enable proper LaTeX rendering."""
    for i, letter in enumerate(consts.GREEK_LETT_LIST_1):
        if 'eta' in s:
            s = _greekReplace(s, 'eta', '\\eta')
        if letter in s:
            s = _greekReplace(s, letter, consts.GREEK_LETT_LIST_2[i])
            # if s[s.find(letter+'_')-1] !='\\':
            #    s = s.replace(letter+'_',GREEK_LETT_LIST_2[i]+'_')
    return s


def _greekReplace(s: str, sub: str, repl: str) -> str:
    """Auxiliary function for _greekPrependify()."""
    # if find_index is not minus1 we have found at least one match for the substring
    find_index = s.find(sub)
    # loop util we find no (more) match
    while find_index != -1:
        if find_index == 0 or (s[find_index - 1] != '\\' and not s[find_index - 1].isalpha()):
            if sub != 'eta':
                s = s[:find_index] + repl + s[find_index + len(sub):]
            else:
                if s[find_index - 1] != 'b' and s[find_index - 1] != 'z':
                    if s[find_index - 1] != 'h':
                        s = s[:find_index] + repl + s[find_index + len(sub):]
                    elif s[find_index - 1] == 'h':
                        if s[find_index - 2] != 't' and s[find_index - 2] != 'T':
                            s = s[:find_index] + repl + s[find_index + len(sub):]
        # find + 1 means we start at the last match start index + 1
        find_index = s.find(sub, find_index + 1)
    return s


def _doubleUnderscorify(s: str) -> str:
    """Set underscores in expressions which need two indices to enable proper LaTeX rendering."""
    ind_list = [kk for kk, char in enumerate(s)
                if char == '_' and s[kk + 1] != '{']
    if len(ind_list) == 0:
        return s
    else:
        index_MinCharLength = 1
        index_MaxCharLength_init = 20
        s_list = list(s)

        for ind in ind_list:
            ind_diff = len(s_list) - 1 - ind
            if ind_diff > 5:
                index_MaxCharLength = min(index_MaxCharLength_init, ind_diff - 5)
                # the following requires that indices consist of 1 or 2 charcter(s) only
                for nn in range(4 + index_MinCharLength, 5 + index_MaxCharLength):
                    if s_list[ind + nn] == '}' and s_list[ind + nn + 1] != '}':
                        s_list[ind] = '_{'
                        s_list[ind + nn] = '}}'
                        break

    return ''.join(s_list)


def _count_sig_decimals(digits: str, maximum: Optional[int] = 7) -> int:
    """Return the number of significant decimals of the input digit string (up to a maximum of 7)."""
    _, _, fractional = digits.partition(".")

    if fractional:
        return min(len(fractional), maximum)
    else:
        return 0


def _format_advanced_option(optionName: str, inputValue, initValues, extraParam=None, extraParam2=None):
    """Check if the user-specified values are within valid range (appropriate subfunctions are called depending on the parameter).

    parameters for slider widgets return list of length 5 as [value, min, max, step, fixed]

    parameters for boolean, dropbox, or input fields return list of lenght two as [value, fixed]

    values is the initial value, (min,max,step) are for sliders, and fixed is a boolean that indciates if the parameter is fixed or the widget should be displayed

    """
    if optionName == 'initialState':
        (allReactants, _) = extraParam
        #fixSumTo1 = extraParam2[0] # until we have better information, all views should sum to 1, then use system size to scale
        fixSumTo1 = True
        idleReactant = extraParam2[1]
        initialState = {}
        # handle initialState dictionary (either convert or generate a default one)
        if inputValue is not None:
            for reactant in sorted(inputValue.keys(), key=str):
                pop = inputValue[reactant]
                initPop = initValues.get(reactant) if initValues is not None else None

                # Convert string into SymPy symbol
                initialState[parse_latex(reactant)] = _parse_input_keyword_for_numeric_widgets(
                    inputValue=pop,
                    defaultValueRangeStep=[defaults.MuMoTdefault._agents,
                                           defaults.MuMoTdefault._agentsLimits[0],
                                           defaults.MuMoTdefault._agentsLimits[1],
                                           defaults.MuMoTdefault._agentsStep],
                    initValueRangeStep=initPop,
                    validRange=(0.0, 1.0) if fixSumTo1 else (0, float("inf")))
                fixedBool = True
        else:
            first = True
            initValuesSympy = ({parse_latex(reactant): pop
                                for reactant, pop in initValues.items()}
                               if initValues is not None else {})
            for reactant in sorted(allReactants, key=str):
                defaultV = defaults.MuMoTdefault._agents if first else 0
                first = False
                initialState[reactant] = _parse_input_keyword_for_numeric_widgets(
                    inputValue=None,
                    defaultValueRangeStep=[defaultV,
                                           defaults.MuMoTdefault._agentsLimits[0],
                                           defaults.MuMoTdefault._agentsLimits[1],
                                           defaults.MuMoTdefault._agentsStep],
                    initValueRangeStep=initValuesSympy.get(reactant),
                    validRange=(0.0, 1.0) if fixSumTo1 else (0, float("inf")))
                fixedBool = False

        # Check if the initialState values are valid
        if fixSumTo1:
            sumValues = sum([initialState[reactant][0] for reactant in allReactants])
            minStep = min([initialState[reactant][3] for reactant in allReactants])
            
            # first thing setting the values of the idleReactant
            if idleReactant is not None: 
                idleValue = initialState[idleReactant][0]
                if idleValue > 1: 
                    wrn_msg = f"WARNING! the initial value of reactant {idleReactant} has been changed to {new_val}\n"
                    warn(wrn_msg, exceptions.MuMoTWarning)
                    initialState[idleReactant][0] = new_val
                # the idleValue have range min-max reset to [0,1]
                initialState[idleReactant][1] = 0
                initialState[idleReactant][2] = 1
                initialState[idleReactant][3] = minStep
            else:
                idleValue = 0
            for reactant in sorted(allReactants, key=str):
                if reactant not in allReactants:
                    error_msg = (f"Reactant '{reactant}' does not exist in this model.\n"
                                 f"Valid reactants are {allReactants}. Please, correct the value and retry.")
                    raise exceptions.MuMoTValueError(error_msg)
    
                # check if the proportions sum to 1
                if reactant != idleReactant:
                    pop = initialState[reactant]
                    # modify (if necessary) the initial value
                    if sumValues > 1:
                        new_val = max(0, pop[0] + (1 - sumValues))
                        if not _almostEqual(pop[0], new_val):
                            wrn_msg = f"WARNING! the initial value of reactant {reactant} has been changed to {new_val}\n"
                            warn(wrn_msg, exceptions.MuMoTWarning)
                            sumValues -= pop[0]
                            sumValues += new_val
                            initialState[reactant][0] = new_val
                    # modify (if necessary) min-max
                    if idleReactant is not None:
                        pop = initialState[reactant]
                        sumNorm = sumValues if sumValues <= 1 else 1
                        if pop[2] > (1 - sumNorm + pop[0] + idleValue):  # max
                            if pop[1] > (1 - sumNorm + pop[0] + idleValue):  # min
                                initialState[reactant][1] = (1 - sumNorm + pop[0] + idleValue)
                            initialState[reactant][2] = (1 - sumNorm + pop[0] + idleValue)
                        if pop[1] > (1 - sumNorm + pop[0]):  # min
                            initialState[reactant][1] = (1 - sumNorm + pop[0])
                        # initialState[reactant][3] = minStep
            if not _almostEqual(sumValues, 1):
                reactantToFix = sorted(allReactants, key=str)[0] if idleReactant is None else idleReactant
                new_val = 1 - sum([initialState[reactant][0]
                                   for reactant in allReactants
                                   if reactant != reactantToFix])
                wrn_msg = f"WARNING! the initial value of reactant {reactantToFix} has been changed to {new_val}\n"
                warn(wrn_msg, exceptions.MuMoTWarning)
                initialState[reactantToFix][0] = new_val
        return [initialState, fixedBool]
        # print("Initial State is " + str(initialState))
    if optionName == 'maxTime':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=([2, 0.1, 3, 0.1]
                                   if extraParam == 'asNoise'
                                   else [defaults.MuMoTdefault._maxTime,
                                         defaults.MuMoTdefault._timeLimits[0],
                                         defaults.MuMoTdefault._timeLimits[1],
                                         defaults.MuMoTdefault._timeStep]),
            initValueRangeStep=initValues,
            validRange=(0, float("inf")))
    if optionName == 'randomSeed':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=np.random.randint(consts.MAX_RANDOM_SEED),
            initValueRangeStep=initValues,
            validRange=(1, consts.MAX_RANDOM_SEED), onlyValue=True)
    if optionName == 'motionCorrelatedness':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=[0.5, 0.0, 1.0, 0.05],
            initValueRangeStep=initValues,
            validRange=(0, 1))
    if optionName == 'particleSpeed':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=[0.01, 0.0, 0.1, 0.005],
            initValueRangeStep=initValues,
            validRange=(0, 1))

    if optionName == 'timestepSize':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=[1, 0.01, 1, 0.01],
            initValueRangeStep=initValues,
            validRange=(0, float("inf")))

    if optionName == 'netType':
        # check validity of the network type or init to default
        if inputValue is not None:
            decodedNetType = utils._decodeNetworkTypeFromString(inputValue)
            if decodedNetType is None:  # terminating the process if the input argument is wrong
                error_msg = (f"The specified value for netType ={inputValue} is not valid. \n"
                             "Accepted values are: 'full',  'erdos-renyi', 'barabasi-albert', and 'dynamic'.")
                raise exceptions.MuMoTValueError(error_msg)

            return [inputValue, True]
        else:
            decodedNetType = utils._decodeNetworkTypeFromString(initValues) if initValues is not None else None
            if decodedNetType is not None:  # assigning the init value only if it's a valid value
                return [initValues, False]
            else:
                return ['full', False]  # as default netType is set to 'full'
    # @todo: avoid that these value will be overwritten by _update_net_params()
    if optionName == 'netParam':
        netType = extraParam
        systemSize = extraParam2
        # if netType is not fixed, netParam cannot be fixed
        if (not netType[-1]) and inputValue is not None:
            error_msg = ("If netType is not fixed, netParam cannot be fixed. "
                         "Either leave free to widget the 'netParam' or fix the 'netType'.")
            raise exceptions.MuMoTValueError(error_msg)
        # check if netParam range is valid or set the correct default range (systemSize is necessary)
        if utils._decodeNetworkTypeFromString(netType[0]) == consts.NetworkType.FULLY_CONNECTED:
            return [0, 0, 0, False]
        elif utils._decodeNetworkTypeFromString(netType[0]) == consts.NetworkType.ERSOS_RENYI:
            return _parse_input_keyword_for_numeric_widgets(
                inputValue=inputValue,
                defaultValueRangeStep=[0.1, 0.1, 1, 0.1],
                initValueRangeStep=initValues,
                validRange=(0.1, 1.0))
        elif utils._decodeNetworkTypeFromString(netType[0]) == consts.NetworkType.BARABASI_ALBERT:
            maxEdges = systemSize - 1
            return _parse_input_keyword_for_numeric_widgets(
                inputValue=inputValue,
                defaultValueRangeStep=[min(maxEdges, 3), 1, maxEdges, 1],
                initValueRangeStep=initValues,
                validRange=(1, maxEdges))
        elif utils._decodeNetworkTypeFromString(netType[0]) == consts.NetworkType.SPACE:
            pass  # method is not implemented
        elif utils._decodeNetworkTypeFromString(netType[0]) == consts.NetworkType.DYNAMIC:
            return _parse_input_keyword_for_numeric_widgets(
                inputValue=inputValue,
                defaultValueRangeStep=[0.1, 0.0, 1.0, 0.05],
                initValueRangeStep=initValues,
                validRange=(0, 1.0))
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=[0.5, 0, 1, 0.1],
            initValueRangeStep=initValues,
            validRange=(0, float("inf")))

    if optionName == 'plotProportions':
        return _parse_input_keyword_for_boolean_widgets(
            inputValue=inputValue,
            defaultValue=False,
            initValue=initValues,
            paramNameForErrorMsg=optionName)
    if optionName == 'realtimePlot':
        return _parse_input_keyword_for_boolean_widgets(
            inputValue=inputValue,
            defaultValue=False,
            initValue=initValues,
            paramNameForErrorMsg=optionName)
    if optionName == 'showTrace':
        return _parse_input_keyword_for_boolean_widgets(
            inputValue=inputValue,
            defaultValue=False,
            initValue=initValues,
            paramNameForErrorMsg=optionName)
    if optionName == 'showInteractions':
        return _parse_input_keyword_for_boolean_widgets(
            inputValue=inputValue,
            defaultValue=False,
            initValue=initValues,
            paramNameForErrorMsg=optionName)

    if optionName == 'visualisationType':
        if extraParam is not None:
            if extraParam == 'multiagent':
                validVisualisationTypes = ['evo', 'graph', 'final', 'barplot']
            elif extraParam == "SSA":
                validVisualisationTypes = ['evo', 'final', 'barplot']
            elif extraParam == "multicontroller":
                validVisualisationTypes = ['evo', 'final']
        else:
            validVisualisationTypes = ['evo', 'graph', 'final']
        if inputValue is not None:
            if inputValue not in validVisualisationTypes:  # terminating the process if the input argument is wrong
                errorMsg = (f"The specified value for visualisationType = {inputValue} is not valid.\n"
                            f"Valid values are: {validVisualisationTypes}. Please correct it and retry.")
                raise exceptions.MuMoTValueError(errorMsg)
            return [inputValue, True]
        else:
            if initValues in validVisualisationTypes:
                return [initValues, False]
            else:
                return ['evo', False]  # as default visualisationType is set to 'evo'

    if optionName in ('final_x', 'final_y'):
        reactants_str = [str(reactant) for reactant in sorted(extraParam, key=str)]
        if inputValue is not None:
            inputValue = inputValue.replace('\\', '')
            if inputValue not in reactants_str:
                error_msg = (f"The specified value for {optionName} = {inputValue} is not valid.\n"
                             f"Valid values are the reactants: {reactants_str}. Please correct it and retry.")
                raise exceptions.MuMoTValueError(error_msg)
            else:
                return [inputValue, True]
        else:
            if initValues is not None:
                initValues = initValues.replace('\\', '')
            if initValues in reactants_str:
                return [initValues, False]
            else:
                if optionName == 'final_x' or len(reactants_str) == 1:
                    return [reactants_str[0], False]  # as default final_x is set to the first (sorted) reactant
                else:
                    return [reactants_str[1], False]  # as default final_y is set to the second (sorted) reactant

    if optionName == 'runs':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=([20, 5, 100, 5]
                                   if extraParam == 'asNoise'
                                   else [1, 1, 20, 1]),
            initValueRangeStep=initValues,
            validRange=(1, float("inf")))

    if optionName == 'aggregateResults':
        return _parse_input_keyword_for_boolean_widgets(
            inputValue=inputValue,
            defaultValue=True,
            initValue=initValues,
            paramNameForErrorMsg=optionName)

    if optionName == 'initBifParam':
        return _parse_input_keyword_for_numeric_widgets(
            inputValue=inputValue,
            defaultValueRangeStep=[defaults.MuMoTdefault._initialRateValue,
                                   defaults.MuMoTdefault._rateLimits[0],
                                   defaults.MuMoTdefault._rateLimits[1],
                                   defaults.MuMoTdefault._rateStep],
            initValueRangeStep=initValues,
            validRange=(0, float("inf")))

    return [None, False]  # default output for unknown optionName


def _get_item_from_params_list(params: List[str], targetName: str) -> Optional[str]:
    """Params is a list (rather than a dictionary) and this method is necessary to fetch the value by name."""
    for param in params:
        if param[0] == targetName or param[0].replace('\\', '') == targetName or param[0].replace('_', '_{') + '}' == targetName:
            return param[1]
    return None


def _almostEqual(a: float, b: float) -> bool:
    epsilon = 0.0000001
    return abs(a - b) < epsilon


def _parse_input_keyword_for_numeric_widgets(
        inputValue: Optional[object],
        defaultValueRangeStep: List[object],
        initValueRangeStep: List[object],
        validRange: Optional[List[object]] = None,
        onlyValue: Optional[bool] = False) -> List[object]:
    """Parse an input keyword and set initial range and default values (when the input is a slider-widget).

    Check if the fixed value is not None, otherwise it returns the default value (samewise for ``initRange`` and ``defaultRange``).
    The optional parameter ``validRange`` is use to check if the fixedValue has a usable value.
    If the ``defaultValue`` is out of the ``initRange``, the default value is move to the closest of the initRange extremes.

    Parameters
    ----------
    inputValue : object
        if not ``None`` it indicated the fixed value to use
    defaultValueRangeStep : list of object
        Default set of values in the format ``[val,min,max,step]``
    initValueRangeStep : list of object
        User-specified set of values in the format ``[val,min,max,step]``
    validRange : list of object, optional
        The min and max accepted values ``[min,max]``
    onlyValue : bool, optional
        If ``True`` then ``defaultValueRangeStep`` and ``initValueRangeStep`` are only a single value

    Returns
    -------
    values : list of object
        Contains a list of five items (start-value, min-value, max-value,
        step-size, fixed). if onlyValue, it's only two items (start-value,
        fixed). The item fixed is a bool. If ``True`` the value is fixed (partial
        controller active), if ``False`` the widget will be created.

    """
    outputValues = defaultValueRangeStep if not onlyValue else [defaultValueRangeStep]
    if not onlyValue:
        if initValueRangeStep is not None and getattr(initValueRangeStep, "__getitem__", None) is None:
            error_msg = (f"initValueRangeStep value '{initValueRangeStep}' must be specified in the format [val,min,max,step].\n"
                         "Please, correct the value and retry.")
            raise exceptions.MuMoTValueError(error_msg)
    if inputValue is not None:
        if not isinstance(inputValue, numbers.Number):
            error_msg = (f"Input value '{inputValue}' is not a numeric vaule and must be a number.\n"
                         "Please, correct the value and retry.")
            raise exceptions.MuMoTValueError(error_msg)
        elif validRange and (inputValue < validRange[0] or inputValue > validRange[1]):
            error_msg = (f"Input value '{inputValue}' has raised out-of-range exception. Valid range is {validRange}\n"
                         "Please, correct the value and retry.")
            raise exceptions.MuMoTValueError(error_msg)
        else:
            if onlyValue:
                return [inputValue, True]
            else:
                outputValues[0] = inputValue
                outputValues.append(True)
                # it is not necessary to modify the values [min,max,step] because when last value is True, they should be ignored
                return outputValues

    if initValueRangeStep is not None:
        if onlyValue:
            if validRange and (initValueRangeStep < validRange[0] or initValueRangeStep > validRange[1]):
                error_msg = (f"Invalid init value={initValueRangeStep} has raised out-of-range exception. Valid range is {validRange}\n"
                             "Please, correct the value and retry.")
                raise exceptions.MuMoTValueError(error_msg)
            else:
                outputValues = [initValueRangeStep]
        else:
            if initValueRangeStep[1] > initValueRangeStep[2] or initValueRangeStep[0] < initValueRangeStep[1] or initValueRangeStep[0] > initValueRangeStep[2]:
                error_msg = (f"Invalid init range [val,min,max,step]={initValueRangeStep}. Value must be within min and max values.\n"
                             "Please, correct the value and retry.")
                raise exceptions.MuMoTValueError(error_msg)
            elif validRange and (initValueRangeStep[1] < validRange[0] or initValueRangeStep[2] > validRange[1]):
                error_msg = (f"Invalid init range [val,min,max,step]={initValueRangeStep} has raised out-of-range exception. Valid range is {validRange}\n"
                             "Please, correct the value and retry.")
                raise exceptions.MuMoTValueError(error_msg)
            else:
                outputValues = initValueRangeStep

    outputValues.append(False)
    return outputValues


def _parse_input_keyword_for_boolean_widgets(inputValue, defaultValue, initValue=None, paramNameForErrorMsg=None):
    """Parse an input keyword and set initial range and default values (when the input is a boolean checkbox)
    check if the fixed value is not None, otherwise it returns the default value.

    Parameters
    ----------
    inputValue : object
        If not None it indicates the fixed value to use
    defaultValue : bool
        dafault boolean value

    Returns
    -------
    value : object
        The keyword value; 'fixed' is a boolean.
    fixed : bool
        If True the value is fixed (partial controller active); if False the widget will be created.

    """
    if inputValue is not None:
        if not isinstance(inputValue, bool):  # terminating the process if the input argument is wrong
            paramNameForErrorMsg = f"for {paramNameForErrorMsg} = " if paramNameForErrorMsg else ""
            errorMsg = (f"The specified value {paramNameForErrorMsg}'{inputValue}' is not valid. \n"
                        "The value must be a boolean True/False.")
            raise exceptions.MuMoTValueError(errorMsg)
        return [inputValue, True]
    else:
        if isinstance(initValue, bool):
            return [initValue, False]
        else:
            return [defaultValue, False]


def _process_params(params):
    paramsRet = []
    paramNames, paramValues = zip(*params)
    for name in paramNames:
        # self._paramNames.append(name.replace('\\','')) ## @todo: have to rationalise how LaTeX characters are handled
        if name in ('plotLimits', 'systemSize'):
            paramsRet.append(name)
        else:
            expr = parse_latex(name.replace('\\\\', '\\'))
            atoms = expr.atoms()
            if len(atoms) > 1:
                raise exceptions.MuMoTSyntaxError(f"Non-singleton parameter name in parameter {name}")
            for atom in atoms:
                # parameter name should contain a single atom
                pass
            paramsRet.append(atom)

    return (paramsRet, paramValues)


def _decodeNetworkTypeFromString(netTypeStr: str) -> Optional[consts.NetworkType]:
    # init the network type
    admissibleNetTypes = {'full': consts.NetworkType.FULLY_CONNECTED,
                          'erdos-renyi': consts.NetworkType.ERSOS_RENYI,
                          'barabasi-albert': consts.NetworkType.BARABASI_ALBERT,
                          'dynamic': consts.NetworkType.DYNAMIC}

    if netTypeStr not in admissibleNetTypes:
        raise exceptions.MuMoTValueError(f"ERROR! Invalid network type argument! Valid strings are: {admissibleNetTypes}")
    return admissibleNetTypes.get(netTypeStr, None)


def _encodeNetworkTypeToString(netType: consts.NetworkType) -> Optional[str]:
    # init the network type
    netTypeEncoding = {consts.NetworkType.FULLY_CONNECTED: 'full',
                       consts.NetworkType.ERSOS_RENYI: 'erdos-renyi',
                       consts.NetworkType.BARABASI_ALBERT: 'barabasi-albert',
                       consts.NetworkType.DYNAMIC: 'dynamic'}

    if netType not in netTypeEncoding:
        raise exceptions.MuMoTValueError(f"ERROR! Invalid netTypeEncoding table! Tried to encode network type: {netType}")
    return netTypeEncoding.get(netType, 'none')


def _round_to_1(x):
    """Used for determining significant digits for axes formatting in plots MuMoTstreamView and MuMoTbifurcationView."""
    if x == 0:
        return 1
    return round(x, -int(math.floor(math.log10(abs(x)))))


def _make_autopct(values):
    def my_autopct(pct):
        total = sum(values)
        val = int(round(pct * total / 100.0))
        return '{p:.2f}%  ({v:d})'.format(p=pct, v=val)
    return my_autopct



from typing import Tuple


class MuMoTdefault:
    """Store default parameters."""

    _initialRateValue = 2  # @todo: was 1 (choose initial values sensibly)
    _rateLimits = (0.0, 20.0)  # @todo: choose limit values sensibly
    _rateStep = 0.1  # @todo: choose rate step sensibly

    @staticmethod
    def setRateDefaults(initRate=_initialRateValue,
                        limits: Tuple[float, float] = _rateLimits,
                        step: float = _rateStep) -> None:
        MuMoTdefault._initialRateValue = initRate
        MuMoTdefault._rateLimits = limits
        MuMoTdefault._rateStep = step

    _maxTime = 3
    _timeLimits = (0, 10)
    _timeStep = 0.1
    # _maxTime = 5
    # _timeLimits = (0, 50)
    # _timeStep = 0.5

    @staticmethod
    def setTimeDefaults(initTime=_maxTime, limits=_timeLimits,
                        step=_timeStep) -> None:
        MuMoTdefault._maxTime = initTime
        MuMoTdefault._timeLimits = limits
        MuMoTdefault._timeStep = step

    _agents = 1.0
    _agentsLimits = (0.0, 1.0)
    _agentsStep = 0.01

    @staticmethod
    def setAgentsDefaults(initAgents: float = _agents,
                          limits: Tuple[float, float] = _agentsLimits,
                          step: float = _agentsStep) -> None:
        MuMoTdefault._agents = initAgents
        MuMoTdefault._agentsLimits = limits
        MuMoTdefault._agentsStep = step

    _systemSize = 10
    _systemSizeLimits = (5, 100)
    _systemSizeStep = 1

    @staticmethod
    def setSystemSizeDefaults(initSysSize: int = _systemSize,
                              limits: Tuple[int, int] = _systemSizeLimits,
                              step: int = _systemSizeStep) -> None:
        MuMoTdefault._systemSize = initSysSize
        MuMoTdefault._systemSizeLimits = limits
        MuMoTdefault._systemSizeStep = step

    _plotLimits = 1.0
    _plotLimitsLimits = (0.1, 5.0)
    _plotLimitsStep = 0.1

    @staticmethod
    def setPlotLimitsDefaults(initPlotLimits: float = _plotLimits,
                              limits: Tuple[float, float] = _plotLimitsLimits,
                              step: float = _plotLimitsStep) -> None:
        MuMoTdefault._plotLimits = initPlotLimits
        MuMoTdefault._plotLimitsLimits = limits
        MuMoTdefault._plotLimitsStep = step



import io
import os

from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

pkgname = 'mumot'

version_namespace = {}
here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, pkgname, '_version.py'), encoding='utf8') as f:
    exec(f.read(), {}, version_namespace)


setup(
    name=pkgname,
    description='Multiscale Modelling Tool',
    version=version_namespace['__version__'],
    author='James A. R. Marshall, Andreagiovanni Reina, Thomas Bose',
    author_email='james.marshall@shef.ac.uk',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/DiODeProject/MuMoT',
    packages=find_packages(),
    package_dir={'': '.'},
    license='GPL-3.0',
    install_requires=[
        'antlr4-python3-runtime',
        'graphviz',
        'ipykernel<4.7',  # needed so no duplicate figures when wiggle ipywidgets
        'ipython',
        'ipywidgets',
        'matplotlib',
        'networkx',
        'notebook<5.5',  # needed if using pyzmq < 17
        'pydstool>=0.90.3',  # min version that allows scipy >= 1.0.0 to be used
        'pyzmq<17',  # needed if using tornado < 5
        'scipy',
        'sympy>=1.4',
        'tornado<5'  # needed to avoid errors with older ipykernel
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'nbval',
            'nbdime',
            'jupyter',
        ],
        'docs': [
            'sphinx',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



"""Constants (inc. symbols) used in MuMoT."""

from enum import Enum

from sympy.parsing.latex import parse_latex

MAX_RANDOM_SEED = 2147483647

EMPTYSET_SYMBOL = parse_latex('1')

GREEK_LETT_LIST_1 = ['alpha', 'beta', 'gamma', 'Gamma', 'delta', 'Delta',
                     'epsilon', 'zeta', 'theta', 'Theta', 'iota', 'kappa',
                     'lambda', 'Lambda', 'mu', 'xi', 'Xi', 'pi', 'Pi', 'rho',
                     'sigma', 'Sigma', 'tau', 'upsilon', 'Upsilon', 'phi',
                     'Phi', 'chi', 'psi', 'Psi', 'omega', 'Omega', 'varrho',
                     'vartheta', 'varepsilon', 'varphi']
GREEK_LETT_LIST_2 = ['\\' + GreekLett for GreekLett in GREEK_LETT_LIST_1]
GREEK_LETT_RESERVED_LIST = ['\\eta', '\\nu', '\\Phi', '(\\eta)', '(\\nu)',
                            '(\\Phi)']
GREEK_LETT_RESERVED_LIST_PRINT = ['eta', 'nu', 'Phi']

INITIAL_RATE_VALUE = 0.5
RATE_BOUND = 10.0
RATE_STEP = 0.1

INITIAL_COND_INIT_VAL = 0.0
INITIAL_COND_INIT_BOUND = 1.0

LINE_COLOR_LIST = ['b', 'g', 'r', 'c', 'm', 'y', 'grey', 'orange', 'k']
MULTIPLOT_COLUMNS = 2


class NetworkType(Enum):
    """Enumeration of possible network types."""

    FULLY_CONNECTED = 0
    ERSOS_RENYI = 1
    BARABASI_ALBERT = 2
    SPACE = 3
    DYNAMIC = 4
